#!/usr/bin/env python3
"""A DataSet is basically a sample vector X, and a label vector y, with
some additional information.

The functions in this file provide functionality for extracting datasets
from polygons.
"""


import copy
import itertools as itt

import numpy as np
from sklearn.model_selection import ParameterGrid
from sklearn.preprocessing import  StandardScaler
import sklearn.metrics

from core.polygon import Polygon

###############################################################################
##                             DataSet
###############################################################################
class DataSet:
    """Just a wrapper to hold the samples and labels together.  And maybe
    a normalizer in the feature.
    """

    ## A strategy to provide unique indices for the processed polygons
    _unique_index = 0
    @classmethod
    def _get_unique_index(cls):
        answer = cls._unique_index
        cls._unique_index = cls._unique_index + 1
        return answer
    
    def n_samples(self):
        if(self.XX is None): return 0
        else: return XX.shape[0]
               
    def __init__(self, XX, yy, s_stencil, feature_type, owner_polygon_indices):
        self.XX = XX
        self.yy = yy
        self.zz = None ## yhat
        self.s_stencil = s_stencil
        self.feature_type = feature_type
        self.n_augment = 1 
        self.is_scaled = False
        self.is_multiplied_by_sigma = False
        self.scikit_scaler = None
        self.sigma = -1

        ## This will be a vector with a length the same as rows
        ## of X or length of y. For overy point, it stores an index
        ## for the polygon to which the point belongs.
        self.owner_polygon_indices = owner_polygon_indices
        assert (self.owner_polygon_indices is None) or \
            (self.owner_polygon_indices.shape[0] == self.yy.shape[0])

    def clone_info(self):
        ans = DataSet(None,None,self.s_stencil,self.feature_type, None)
        self.n_augment = self.n_augment 
        self.is_scaled = self.is_scaled 
        self.is_multiplied_by_sigma = self.is_multiplied_by_sigma 
        self.scikit_scaler = self.scikit_scaler
        self.sigma = self.sigma 
        return ans

    def clone_full(self):
        return copy.deepcopy(self)

## TYPES
# 0  coordinates of the neighbourhodd
# 1 turn types of the neighbourhood
# 2 coorndinates of the neighbourhood + angles for closure candidates

N_AUGMENT_COORD  = 8
N_AUGMENT_TURNS   = 2
N_AUGMENT_CLOSURE = 8
def INVALID_COORD_VALUE(s): return -s-1

###############################################################################
##                           Data extraction
###############################################################################
def dataset_create(poly, s_stencil, feature_type):
    """ Create the data set. Get feature_type dynamically.
    """
    creators = {
        0: dataset_from_coords,
        1: dataset_from_turns,
        2: dataset_from_closure,
    }
    return creators[feature_type](poly, s_stencil)
 
#-------------------------------------------------
def dataset_from_coords(poly, s_stencil):
    """Return the sample matrix and the label vector
    corresponding to all the points in the polygon.

    This version uses the original relative coordinate representation.
    The XX vector in this case has the shape (npoints, 2s, 2)
    reshape it to (npoints,4s) before feeding it to the regressor.

    The points are always ordered, such that rotating the tangent 
    by 90 degrees points outside.
    """    

    ## Make sure the polygon is directed in the right order
    assert poly.orientation == 1, \
        "Polygon points ordered in the wrong way. Reverse the order."
    # print("# %s" % poly.orientation)
    
    points = poly.points
    iscorner = np.array(poly.corners, dtype=int)
    n_points = poly.n_points()
    n_sample_points = poly.n_feature_points() ## fix this naming shit
    
    # Create the sample matrix
    XX = np.zeros((n_sample_points, 2*s_stencil, 2), dtype=float)

    # Extract the neighbourhood for every point
    # move the coordinate to the center point
    for isp in range(0, n_sample_points ):
        ip = poly.feature_points[isp]
        stencil_indices = np.arange( ip-s_stencil, ip+s_stencil+1)
        stencil_indices = np.delete( stencil_indices, s_stencil, 0)
        stencil_indices = np.mod( stencil_indices + n_points, n_points )
        XX[isp] = points[ stencil_indices ]
        XX[isp] = XX[isp] - points[ip]

    XX = XX.reshape(n_sample_points, 4*s_stencil)
    owner_polygon_indices = np.zeros(n_sample_points, dtype=int) + DataSet._get_unique_index()
        
    return DataSet(XX, iscorner, s_stencil, 0, owner_polygon_indices)
   
#-------------------------------------------------    
def dataset_from_turns(poly, s_stencil):
    """Return the sample matrix and the label vector
    corresponding to all the points in the polygon.

    This version uses turn policy representation.

    NOT TeSTED.
    """
    n_samples = poly.n_feature_points()
    n_points = poly.n_points()

    # Create the sample matrix
    XX = np.zeros((n_samples, 2*s_stencil+1), dtype=float)

    # Extract the neighbourhood for every point
    for isample in range(0, n_samples ):
        ip = poly.feature_points[ isample ]
        stencil_indices = np.arange( ip-s_stencil, ip+s_stencil+1)
        stencil_indices = np.mod( stencil_indices + n_points, n_points )
        XX[isample] = poly.turn_policies[ stencil_indices ]

    owner_polygon_indices = np.zeros(n_points, dtype=int) + DataSet._get_unique_index()
    
    return DataSet(XX, poly.corners, s_stencil, 1, owner_polygon_indices)

#-------------------------------------------------
def dataset_from_closure(poly, s_stencil):
    """Return the sample matrix and the label vector
    corresponding to all the points in the polygon.

    This version uses the original relative coordinate representation.
    The XX vector in this case has the shape (npoints, 4s+4s+6).
    The extra 6 values are distance, alpha, theta1, theta2, theta3, theta4. 

    The points are always ordered, such that rotating the tangent 
    by 90 degrees points outside.
    """    

    ## Make sure the polygon is alright
    assert(poly.n_closure_candidates() >= 1)
    assert(poly.n_closure_features() == 7) # index candid + vals
    assert(poly.orientation == 1)

    ## First get the neighbourhoods from before.
    dset = dataset_from_coords(poly, s_stencil)
    dset.feature_type = 2
    XX_old = dset.XX
    dset.XX = np.zeros( (poly.n_points(), 8*s_stencil+6) )
    dset.XX[:,:4*s_stencil] = XX_old
    dset.XX[:,4*s_stencil:8*s_stencil] = INVALID_COORD_VALUE(s_stencil)
    dset.XX[:,8*s_stencil:] = poly.closure_data[:,0,1:]

    ## Copy the data from the pair (only the first candidate for now)
    pairs = poly.closure_data[:,0,0].astype(dtype=int)
    mask = pairs >= 0
    dset.XX[mask,4*s_stencil:8*s_stencil] = XX_old[pairs[mask],:]
    
    return dset

#-------------------------------------------------    
def dataset_merge(datasets):
    """ Merges data sets """
    answer = datasets[0].clone_info()
    XX_list = []
    yy_list = []
    zz_list = []
    owner_polygon_indices_list = []
    for d in datasets:
        XX_list.append(d.XX)
        yy_list.append(d.yy)
        zz_list.append(d.zz)
        owner_polygon_indices_list.append(d.owner_polygon_indices)
    answer.XX = np.concatenate(XX_list, axis=0)
    answer.yy = np.concatenate(yy_list, axis=0)
    answer.owner_polygon_indices = np.concatenate(owner_polygon_indices_list, axis=0)    
    if(zz_list[0] is not None): answer.zz = np.concatenate(zz_list, axis=0)

    return answer

    
###############################################################################
##                           Data augmentation
###############################################################################
#-------------------------------------------------
def _should_flip(mirror_x, mirror_y, mirror_bisector):
    should_flip = False
    if(mirror_x): should_flip = not should_flip
    if(mirror_y): should_flip = not should_flip
    if(mirror_bisector): should_flip = not should_flip
    return should_flip
    
    
def _augment_coords(XX, mirror_x, mirror_y, mirror_bisector):
    """ Augements to coordinate stencils.
    There are 8 possible duplicates for each stencil.
    Given four coordinate axis west,north,east and south.
    (x,y)=(west,north)*  (x,y)=(east,north)
    (x,y)=(west,south)   (x,y)=(east,south)*
    (x,y)=(north,west)   (x,y)=(south,west)*
    (x,y)=(north,east)*  (x,y)=(south,east)
    The points are always ordered, such that rotating the tangent 
    by 90 degrees points inside (or outside as long as it is 
    always consistent.) Therefore, cases shown by (*) need to have the
    points in the reverse order.
    """
    ## Get the new points
    XX = np.matrix.copy( XX )
    
    if(mirror_x):
        XX[:,:,0] = -XX[:,:,0]
    if(mirror_y):
        XX[:,:,1] = -XX[:,:,1]
    if(mirror_bisector):
        XX[:,:,[0,1]] = XX[:,:,[1,0]]
        
    if( _should_flip(mirror_x, mirror_y, mirror_bisector) ):
        XX =  np.flip(XX, 1)
        
    return XX

#-------------------------------------------------    
def dataset_augment(dataset):
    augmentors = {
        0: dataset_augment_coords,
        1: dataset_augment_turns,
        2: dataset_augment_closure,
    }
    return augmentors[dataset.feature_type](dataset)

#-------------------------------------------------    
def dataset_augment_coords(dataset):
    """ Augments the data set by considering all duplicate cases.
    This version works for features that are actual coordinate. 

    NOT TeSTED.
    """
    assert(dataset.feature_type == 0 )
    n_augment = N_AUGMENT_COORD

    s_stencil =  dataset.s_stencil
    n_points = dataset.XX.shape[0]
    assert(dataset.XX.shape[1] == 4*s_stencil)
    assert(dataset.yy.shape[0] == n_points)

    XX = dataset.XX.reshape(n_points,s_stencil*2,2)     
    XX_aug = np.zeros((n_points*n_augment,2*s_stencil, 2) )
    yy_aug = np.zeros(n_points* n_augment, dtype=int)
    owner_polygon_indices_aug = np.zeros(n_points* n_augment, dtype=int)

    iaug = 0
    for (mx,my,mbs) in itt.product([True,False],repeat=3):
        # beg = iaug*XX.shape[0]
        # end = (iaug+1)*XX.shape[0]
        step = n_augment
        XX_aug[iaug::step,:,:]=_augment_coords(XX,mx,my,mbs)
        yy_aug[iaug::step] = dataset.yy
        owner_polygon_indices_aug[iaug::step] = dataset.owner_polygon_indices
        iaug = iaug + 1
        
    assert(iaug == n_augment)
    answer = dataset.clone_info()
    answer.XX = XX_aug.reshape(n_points*n_augment, 4*s_stencil)
    answer.yy = yy_aug
    answer.owner_polygon_indices = owner_polygon_indices_aug
    answer.n_augment = n_augment
    return answer

#-------------------------------------------------    
def dataset_augment_turns(dataset):
    """ Augments the data set by considering all duplicate cases.
    This version works for features that are turn types. 
    """
    assert(dataset.feature_type == 1 )
    n_augment = N_AUGMENT_TURNS

    s_stencil =  dataset.s_stencil
    n_samples0 = dataset.XX.shape[0]
    n_features = dataset.XX.shape[1]
    assert(dataset.XX.shape[1] == 2*s_stencil + 1)
    assert(dataset.yy.shape[0] == dataset.XX.shape[0])
    
    XX_aug = np.zeros((n_samples0*n_augment, n_features) )
    yy_aug = np.zeros(n_samples0* n_augment)
    owner_polygon_indices_aug = np.zeros(n_points* n_augment, dtype=int)

    XX_aug[0::n_augment] = dataset.XX
    yy_aug[0::n_augment] = dataset.yy
    owner_polygon_indices_aug[0::n_augment] = dataset.owner_polygon_indices
    XX_aug[1::n_augment] = np.flip(dataset.XX,1)
    yy_aug[1::n_augment] = dataset.yy
    owner_polygon_indices_aug[1::n_augment] = dataset.owner_polygon_indices
    
    answer = dataset.clone_info()
    answer.XX = XX_aug
    answer.yy = yy_aug
    answer.owner_polygon_indices = owner_polygon_indices_aug
    answer.n_augment = n_augment
    return answer

#-------------------------------------------------    
def dataset_augment_closure(dataset):
    """ Augments the data set by considering all duplicate cases.
    This version works for features that are actual coordinate +
    the closure data.

    ASSUMPTION: NO PAIR VALUE FOR PAIR STENCIL IS -1.
    NOT TeSTED.
    """
    ## Get the number of points and stencil size
    assert(dataset.feature_type == 2 )
    n_augment = N_AUGMENT_CLOSURE
    s_stencil =  dataset.s_stencil
    ss = s_stencil
    n_points = dataset.XX.shape[0]
    assert(dataset.XX.shape[1] == 8*s_stencil + 6)
    assert(dataset.yy.shape[0] == n_points)

    ## Some aliases for features
    d_closure = 6
    d_geom = 4*ss
    d_total = 2*d_geom + d_closure
    
    idx_self = np.arange(0,4*ss)
    idx_pair = np.arange(4*ss,8*ss)
    idx_closure = np.arange(8*ss, 8*ss+6)
    idx_distance = 8*ss

    ## Get each neighbourhood and the distances
    XX = dataset.XX
    XXself = XX[:,idx_self].reshape(n_points,d_geom/2,2)
    XXpair = XX[:,idx_pair].reshape(n_points,d_geom/2,2)
    has_nopair = XX[:, idx_distance ] < 0
    

    ## Create the augemented data set
    XX_aug = np.zeros((n_points*n_augment,d_total) )
    yy_aug = np.zeros(n_points* n_augment, dtype=int)
    owner_polygon_indices_aug = np.zeros(n_points*n_augment,dtype=int)

    ##
    ## Fill the augmented data set
    ##
    iaug = 0
    step = n_augment
    for (mx,my,mbs) in itt.product( *itt.repeat([True,False],3) ):
        should_flip = _should_flip( mx, my, mbs )
        ## Augment self stenicl
        XX_aug_sub = (
            _augment_coords(XXself,mx,my,mbs).reshape(n_points, d_geom) )
        XX_aug[iaug::step,idx_self] =  XX_aug_sub
        ## Augment pair stencil
        XX_aug_sub = (
            _augment_coords(XXpair,mx,my,mbs).reshape(n_points, d_geom) )
        XX_aug_sub[has_nopair] = INVALID_COORD_VALUE(ss)
        XX_aug[iaug::step,idx_pair] = XX_aug_sub 
        ## Augment the angles
        if( should_flip ):
            XX_aug[iaug::step,idx_closure] = (
                XX[:,  idx_closure[ [0,1,3,2,5,4] ] ] )
            pass
        else:
            XX_aug[iaug::step,idx_closure] = (
                XX[:,  idx_closure ] ) 
        ## Augment the labels    
        yy_aug[iaug::step] = dataset.yy
        owner_polygon_indices_aug[iaug::step] = dataset.owner_polygon_indices
        iaug = iaug + 1    
    assert(iaug == n_augment)

    
    ##
    ## Clone and return
    ##
    answer = dataset.clone_info()
    answer.XX = XX_aug
    answer.yy = yy_aug
    answer.owner_polygon_indices = owner_polygon_indices_aug
    answer.n_augment = n_augment
    return answer

###############################################################################
##                           Data processing
###############################################################################
#-------------------------------------------------    
def dataset_shuffle(d):
    n_points = d.XX.shape[0]
    indices = np.arange(n_points)
    np.random.shuffle(indices)
    d.XX = d.XX[indices]
    d.yy = d.yy[indices]

#-------------------------------------------------    
def dataset_create_scaler(dataset, with_mean=True, with_std=True):
    assert( dataset.scikit_scaler is None ) 
    dataset.scikit_scaler = StandardScaler(True,with_mean,with_std)
    dataset.scikit_scaler.fit(dataset.XX)

#-------------------------------------------------    
def dataset_scale(dataset):
    assert(not dataset.is_scaled)
    dataset.XX = dataset.scikit_scaler.transform(dataset.XX)
    dataset.is_scaled = True

#-------------------------------------------------    
def dataset_unscale(dataset):
    assert(dataset.is_scaled)    
    dataset.XX = dataset.scikit_scaler.inverse_transform(dataset.XX)
    dataset.is_scaled = False
    
#-------------------------------------------------    
def dataset_multiply_gaussian(dataset,sigma):
    assert(not dataset.is_multiplied_by_sigma)

    ss = dataset.s_stencil
    cols1 = None
    hasnopair = None

    ## When pairs are present, they are only multiplied by gaussian
    ## when they exist.
    if(dataset.feature_type == 2):
        hasnopair = dataset.XX[:, 8*ss] < 0
        n_hasnopair = np.sum(hasnopair)
        invalid_coord_value = np.zeros(
            (n_hasnopair,8*ss+6))+INVALID_COORD_VALUE(ss)
        if(dataset.scikit_scaler is not None):
            invalid_coord_value = (
                dataset.scikit_scaler.transform(invalid_coord_value) )

    ## Loop over features
    for i in range(0,ss):
        if(dataset.feature_type == 0 ):
            cols1 = [2*ss-2*i-2,2*ss-2*i-1,
                    2*ss+2*i, 2*ss+2*i+1]
        elif(dataset.feature_type == 1 ):
            cols1 = [ss-i-1, ss+i+1]
        elif(dataset.feature_type == 2 ):
            cols1 = [2*ss-2*i-2 ,2*ss-2*i-1,
                     2*ss+2*i   ,2*ss+2*i+1,
                     6*ss-2*i-2 ,6*ss-2*i-1,
                      6*ss+2*i   ,6*ss+2*i+1,]

        coeff = np.exp(-float(i+1)*(i+1)/2./sigma/sigma)
        dataset.XX[:,cols1] = dataset.XX[:,cols1] * coeff

    ## Restore values for those who did not have neighbours
    if(hasnopair is not None):
        dataset.XX[hasnopair,4*ss:8*ss] = invalid_coord_value[:,4*ss:8*ss]
    
    dataset.sigma = sigma
    dataset.is_multiplied_by_sigma = True

###############################################################################
##                        Inverse Augmentation!!
###############################################################################
#-------------------------------------------------    
def _threshold(mat, thr=0, vmin=-1, vmax=1):
    """
    Threshold probablity values by clamping them to vmin and vmax.
    """
    ans = mat.copy()
    vmin_entries = ans <= thr
    ans.fill(vmax)
    ans[vmin_entries] = vmin
    return ans

#-------------------------------------------------
class _ColumnChooser:
    def __init__(self, col_index):
        self.col_index  = col_index
    def choose(self, mat):
        return mat[:, self.col_index]
    
#-------------------------------------------------    
def unaugment_scores(yy, operation, n_chunk_mems):
    """
    Apply a certain operation to every ``n_chunk_mems'' inside a numpy
    ndarray.
    """
    vec2 = yy.copy().reshape(-1,n_chunk_mems)
    vec2 = operation(vec2,1)
    return vec2.flatten()

#-------------------------------------------------    
def threshold_scores(mat, thr=0, vmin=-1, vmax=1):
    """
    Threshold probablity values by clamping them to vmin and vmax.
    """
    ans = mat.copy()
    vmin_entries = ans <= thr
    ans.fill(vmax)
    ans[vmin_entries] = vmin
    return ans

###############################################################################
##                               Metrics
###############################################################################
#-------------------------------------------------
def eval_fscore(pres, recall,beta=1):
    """
    Find the fscore as a function of precision and recall.
    """
    nom = (1+beta*beta)* np.multiply(pres, recall)
    denom = beta*beta*recall+pres
    ans = np.divide(nom,denom)
    ans = np.nan_to_num(ans, copy=True)
    return ans

#-------------------------------------------------    
def best_fscore(dset, betas, ave=np.amax):

    ## Gather if required
    yy = unaugment_scores(dset.yy, ave, dset.n_augment)
    zz = unaugment_scores(dset.zz, ave, dset.n_augment)

    ## Find precision and recall     
    precision, recall, thresholds = (
        sklearn.metrics.precision_recall_curve( yy, zz, pos_label=1) )
    
    ## Find the best fscore
    ans = np.zeros((len(betas),3),dtype=float)
    for i,b in enumerate(betas):
        fscore = eval_fscore(precision, recall, float(b))[:-1]
        idx = np.argmax(fscore)
        ans[i,0] = fscore[idx]
        ans[i,1] = precision[idx]
        ans[i,2] = recall[idx]
    
    return ans

#-------------------------------------------------    
def plotty_fscore(y_true, y_proba, beta, n_augment, gather_op):
    
    ## Gather if required
    y_true = unaugment_scores(y_true, gather_op, n_augment)
    y_proba = unaugment_scores(y_proba, gather_op,  n_augment)

    ## Find precision and recall     
    precision, recall, thresholds = (
        sklearn.metrics.precision_recall_curve( y_true, y_proba, pos_label=1) )
    fscore = eval_fscore(precision, recall, beta )

    ## duplicate for plotting purposes
    thresholds = np.append([np.append([-1], thresholds)],
                           [np.append(thresholds, [1])],0)
    thresholds = thresholds.transpose().flatten()
    
    precision = np.append([precision],[precision],0).transpose().flatten()
    recall = np.append([recall],[recall],0).transpose().flatten()
    fscore = np.append([fscore],[fscore],0).transpose().flatten()
    
    return (thresholds, fscore, precision, recall)


###############################################################################
##                              Create a data set
###############################################################################

def make_dataset(polygons,
                 feature_type,
                 s_stencil,
                 sigma,
                 should_mirror,
                 should_scale): # either tuple, or (scikit_scaler)
    """ Given a set of polygons, extracts a data set out of them.
    Also returns the group id assigned to every polygon.
    """
    group_ids = []
    
    ##
    ## create seperate data sets and merge them to one
    ## Feature type is important here.
    ## ONLY ft=0 and 2 supported for now.
    ##
    dset_polys = []
    for pp in polygons:
        dset = dataset_create(pp, s_stencil, feature_type)
        dset_polys.append(dset)
        group_ids.append(dset.owner_polygon_indices[0])
    dset = dataset_merge(dset_polys) 

    
    ##
    ## preprocessing
    ##
    
    
    ## Mirror if needed
    ## Feature type is important here. ONLY ft=0 now
    if ( should_mirror ):
        dset = dataset_augment(dset)

    ## Scaling
    ## 1- should scale is (False, ..., ...)
    ## 2- should scale is (True, scale_mean, scale_standard_deviation)
    ## 3- should scale is (scikit_scale, ...)
    if(should_scale is None):
        pass
    elif( isinstance(should_scale, StandardScaler) ):
        dset.scikit_scaler = should_scale
        dataset_scale(dset)
    elif( should_scale[0] == True ): # maybe assert that is a tuple
        dataset_create_scaler(dset,
                              should_scale[1], should_scale[2])
        dataset_scale(dset)
    else:
        pass
        
    ## Multiply by sigma    
    if ( sigma > 0 ):
        dataset_multiply_gaussian(dset, sigma)
        

    return (dset, group_ids)
