#!/usr/bin/env python3


import core.png_bw_tools as bwt
import core.png_color_tools as mct
from core.polygon import Polygon
from core.classifier import Classifier
import core.fitter_options as fitter
from  core.defs import *
from core.extract_multicolor import *

import sys, os, subprocess, shutil, glob
import numpy as np
import imageio

class Manager:
    """All the flow of the application is managed by this guy.  It is
completely decoupled from the GUI, and can be used on its own. The GUI
will call the appropriate functions from the Manager.

    """

    def __init__(self):
        self.input_png_address = ''
        self.fitter_address = ''
        self.tmp_address = ''
        ###
        self.image_raw = None
        self.image = None
        self.image_options = None
        ###
        self.polygon = None
        ###
        self.svg_address = None
        ###
        self.observers = []
        ###
        self.to_view = VIEW_NONE
        ###
        self.svg_bw_addr = None
        self.svg_color_addr = None
        self.models = {}
        self.view_colors = 0
        self.is_multicolor = 0
        self.read_models()
        
        if not os.path.exists(TEMP_DIR):
            os.makedirs(TEMP_DIR)
        

    def read_models(self):
        for reso in [8, 16, 32, 64, 128, 256]:
            self.models[reso] = Classifier()
            self.models[reso].read_from_file(CLASSIFIER_DIR[reso])

    def add_observer(self, obs):
        self.observers.append(obs)

    def call_observers(self, method):
        for obs in self.observers:
            func = getattr(obs, method)
            func(self)

    ## ======================================
    ##   CHANGING WHAT TO VIEW
    ## ======================================
    def change_view_mode(self, mode):
        self.to_view = mode
        self.call_observers('observe_image_changed')
        
    ## ======================================
    ##   IMAGE MODIFICATION ROUTINES
    ## ======================================
    def read_image(self, address, is_color):
        self.input_png_address = address
        print( "Reading {}..".format(address) )
        self.is_multicolor=bool(is_color)
        if(self.is_multicolor):
            self.image_raw = mct.read_img(self.input_png_address)
        else:
            self.image_raw = bwt.read_img(self.input_png_address)
        self.reset_image()
        
    def reset_image(self):
        if(self.is_multicolor):
            self.image = mct.rescale(self.image_raw)
        else:
            self.image = bwt.threshold(self.image_raw)
            if(not self.image[0,0]): self.invert_image()
        self.change_view_mode( VIEW_PNG )

    def invert_image(self):
        if(not self.is_multicolor):
            self.image = ~self.image
            self.change_view_mode( VIEW_PNG )

    def crop_image(self):
        if(not self.is_multicolor):
            self.image = bwt.crop_bbox(self.image)
            self.change_view_mode( VIEW_PNG )

    def extract_boundary(self):
        if(not self.is_multicolor):
            if( self.image is  None): return
            verts = bwt.extract_polygon(self.image)
            self.polygon = Polygon()
            self.polygon.init_from_verts( verts )
            self.change_view_mode( VIEW_POLY )

    def run_classifier(self):
        if( self.polygon is  None): return
        if( self.is_multicolor): return 
        reso = self.polygon.get_reso()
        self.models[reso].predict(self.polygon)
        self.change_view_mode( VIEW_SCORES )

    def save_image(self, filename):
        if self.svg_address is None:
            print('Vectorize something first!')
            return

        try:
            shutil.copyfile(self.svg_address, filename)
        except IOError:
            print('Failed to copy vectorized image to %s'%filename)

    def run_fitter(self):
        try:
            shutil.rmtree(TEMP_DIR)
            os.makedirs(TEMP_DIR)
        except:
            pass
        
        if( self.image is None): return
        if( self.is_multicolor):
            self.run_fitter_mc()
        else:
            self.run_fitter_bc()
        
    def run_fitter_bc(self):
        self.extract_boundary()
        self.run_classifier()
        if( Y_PREDICT not in self.polygon.point_data): return
        if( Y_PROBA   not in self.polygon.point_data): return
        reso = self.polygon.get_reso()
        self.svg_address = fitter.run_fitter(self.polygon, reso)
        self.change_view_mode( VIEW_SVG )

    
    def run_fitter_mc(self):
        print('Processing image...')
        # Let's first clear all the temporary data
        if os.path.exists(MULTIPOLYGON_DIR):
            shutil.rmtree(MULTIPOLYGON_DIR, ignore_errors=True)

        # Saving rescaled image
        imageio.imwrite(EXPANDED_PNG_DIR, self.image)
        
        # Creating directories
        if not os.path.exists(MULTIPOLYGON_DIR):
            os.makedirs(MULTIPOLYGON_DIR)

        print('Extracting polygons...')
        # Extracing all polygons in all_polys folder
        extract_multicolor(EXPANDED_PNG_DIR, MULTIPOLYGON_DIR)

        print('Finding corners...')
        # Running classifier on each subpolygon
        i = 0
        max_res = -1
        while True:
            # Making sure polygon exists
            model_path = '%s/%s.poly'%(MULTIPOLYGON_DIR, i)
            try:
                with open(model_path, 'r') as ff: pass
            except: 
                print('Finished classifier on %d polygons'%(i+1))
                break
            i = i + 1

            # Reading
            polygon = Polygon()
            polygon.read_dot_poly(model_path)
            if (polygon.orientation == -1): polygon.mirror_points(False, False, False, True)

            # Finding classifier resolution to use
            poly_min = [2**14, 2**14]
            poly_max = [-1, -1]
            for point in polygon.points:
                if point[0] < poly_min[0]: poly_min[0] = point[0]
                if point[1] < poly_min[1]: poly_min[1] = point[1]
                if point[0] > poly_max[0]: poly_max[0] = point[0]
                if point[1] > poly_max[1]: poly_max[1] = point[1]

            poly_size = max(poly_max[0] - poly_min[0], poly_max[1] - poly_min[1])

            if poly_size < 10: 
                resolution = 8
            elif poly_size < (16+32)/2:
                resolution = 16
            elif poly_size < (32+64)/2:
                resolution = 32
            elif poly_size < (64+128)/2:
                resolution = 64
            elif poly_size < (128+256)/2:
                resolution = 128
            else:
                resolution = 128
            max_res = max(max_res, resolution)

            self.models[resolution].predict(polygon)
            polygon.write_dot_poly(model_path)

        # Running fitter
        print('Fitting...')
        self.svg_bw_addr, self.svg_color_addr = fitter.run_fitter_multi(max_res)
        self.svg_address = self.svg_bw_addr
        self.view_colors = 0
        self.change_view_mode( VIEW_SVG )        

        print('Finished.')

        # Removing log files
        for f in glob.glob("*.log"):
            os.remove(f)

    def toggle_view_colors(self):
        if(not self.is_multicolor): return
        if self.view_colors == 0:
            self.svg_address = self.svg_color_addr
            self.view_colors = 1
        else:
            self.svg_address = self.svg_bw_addr
            self.view_colors = 0
        self.change_view_mode( VIEW_SVG )
