#!/usr/bin/env python3
"""png_bw_tools.py

Basic pre-processing for png images.
Convert png to binary colors.
Scale png.
Trace the polygon boundary out of the png image.

NOTE: NumPy indexing is like matrix indexing An image is processed
       exactly like a matrix.  READ THIS
       http://spencermortensen.com/articles/bezier-circle/

"""

import sys
import copy

import imageio
import numpy as np

######################################################################
##                    Remove alpha mask
######################################################################
def remove_alpha(img):
    ## If alpha exists, remove it
    if( (img.ndim == 3) and (img.shape[2] == 4)):
        img2 = np.delete(img, 3, 2)
        mask = img[:,:,3] < 0.1
        img2[mask] = 255
        return img2
    else:
        return img

######################################################################
##                  Convert to binary image
######################################################################
def grayscale(img, ww=(0.299, 0.587, 0.114)):
    if( img.ndim == 3):
        img = remove_alpha(img)
        fimg = img.astype('float')
        gray = ww[0]*fimg[:,:,0] + ww[1]*fimg[:,:,1] + ww[2]*fimg[:,:,2]
        return gray.astype('uint8')
    else:
        return img


######################################################################
##                  Threshold the image
######################################################################
def threshold(img, thr=125, should_invert=False):
    """Black is converted to 0 normally, i.e. value < thr if
    selfhould_invert == True, white will be converted to 0, i.e.
    value > thr

    """
    ## If the image is in color make it grayscale
    if (img.ndim > 2): img = grayscale(img)
    if(not should_invert): return (img > thr)
    else:                  return (img <= thr)

######################################################################
##                      Finding bounding box
######################################################################
def get_bbox(img, thr=125):
    ## only for graysale
    assert(img.ndim==2)

    ## False is considered inside the polygon
    ## Pretty strange(!) I know, maybe I should change it.
    a = np.where(img == False)
    bbox = (np.min(a[0]), np.max(a[0])), (np.min(a[1]), np.max(a[1]))
    return bbox

######################################################################
##                          Cropping
######################################################################
def crop(img, rows=(0,-1), cols=(0,-1)):
    return img[rows[0]:rows[1] , cols[0]:cols[1]]

def crop_bbox(img):
    bbox = get_bbox(img)
    img2 = crop(img, (bbox[0][0], bbox[0][1]+1),
               (bbox[1][0], bbox[1][1]+1) )
    return img2


######################################################################
##                     Inverse nearest neighbour
######################################################################
def inverse_nn(img, scale):
    if(scale==1): return img

    shape = list(img.shape)
    shape[0] = shape[0] / scale
    shape[1] = shape[1] / scale

    ans = np.zeros(shape, img.dtype)

    for ii in range(0, shape[0]):
        for jj in range(0, shape[1]):
            
            JJ = jj*scale;
            II = ii*scale

            pixel = crop(img, (II, II+scale), (JJ, JJ+scale))
            maxval = np.amax( pixel )
            ans[ii,jj] = maxval

    return ans         

######################################################################
##                     Read the image
######################################################################
def read_img(fname):
    return imageio.imread(fname)
    

######################################################################
##                     Extract boundary polygon
######################################################################

## Cell and vertex numbering is such that:
## 
##  v:(i,j)------v:(i,j+1)
##  |               |
##  | c:(i,j)       |
##  |               |
##  v:(i+1,j)----v:(i+1,j+1)
## Also, 0,0 is the top left corner. 
def extract_polygon(img):
    isin = (img == 0)
    map_vertex = {}

    ## -------------------
    ##  Loop over all pixels.
    ##  For every pixel which is inside, get all its edges (CCW).
    ##  If an edge lies on the boundary, put it in the map.
    ## -------------------
    h = img.shape[0]
    w = img.shape[1]
    for i in range(0, h):
        for j in range(0, w):

            if( not isin[i,j] ): continue

            ## Vertices of edges (left, down, right, up)
            verts_of_edges = [(i,j), (i+1,j),
                              (i+1,j), (i+1,j+1),
                              (i+1,j+1), (i,j+1),
                              (i,j+1), (i,j)]


            ## Are neighbours inside (left, down, right, up)
            isneighin = [False] * 4
            isneighin[0] = (False if j==0   else   isin[(i)  ,j-1] )
            isneighin[1] = (False if i==h-1 else   isin[(i+1),j  ] )
            isneighin[2] = (False if j==w-1 else   isin[(i)  ,j+1] )
            isneighin[3] = (False if i==0   else   isin[(i-1), j ] )

            # Add the boundary edges to the map
            for e in range(0,4):
                ##  inner edge
                if (isneighin[e]): continue
                ##  The key to be inserted
                key =  verts_of_edges[2*e]
                val =  verts_of_edges[2*e+1]
                if (key in map_vertex):  raise TypeError("Non manifold")
                map_vertex[key] = val

    ## -------------------
    ##  Extract the edges
    ## -------------------
    vertices = []
    vert, next_vert = map_vertex.popitem()
    first_vert = vert

    while (next_vert != first_vert):
        ## Add vertex
        vertices.append(vert)

        ## find the next vertex
        if( not (next_vert in map_vertex) ):
            raise TypeError("Failed to find next vertex")
        vert = next_vert
        next_vert = map_vertex[vert]
    ## Don't miss the last vertex!    
    vertices.append(vert)

    ans = np.array(vertices, dtype=int)
    ans[:,[0,1]] = ans[:,[1,0]]
    return ans
