#!/usr/bin/env python3

######################################################################
##                          WHAT SHOULD BE DONE
######################################################################

## polygon.py

## Class Polygon.
## A polygon should be able to read a polygon, and the corners
## from .txt files.

import copy
import os

import numpy as np
import sklearn.preprocessing as sklpr

from core.defs import *
from core.line_by_line_reader import LineByLineReader

import matplotlib
matplotlib.use(MPL_BACKEND)
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import matplotlib.colors as mcolors

_cmap_of_choice = plt.get_cmap('rainbow')

## POlygon is not aware what the data it is storing anymore
## THis is only for the older io functions.
CORNER_SCORES = 'corner_scores'

######################################################################
##                            Polygon
######################################################################
class Polygon:

    ## -------------------- Constructor
    def init(self):
        self.name = ''
        self.points = np.zeros((0,0))
        ## n_p * n_rays * len([ip, metrics])
        self.closure_data = np.zeros((0,0,0))
        self.feature_points = np.zeros(0)
        self.xcorner2sample = np.zeros(0)        
        self.corners = np.zeros(0)
        self.point_data = {}
        self.orientation = int(-1)
        self.turn_policies = np.zeros(0)
        self.resolution = -1

    def __str__(self):
        return self.name
    
    __repr__ = __str__
        
    def __init__(self):
        self.init()

    def update_orientation(self):
        """
        Orientation is set such that ori > 0, means \int ydx > 0
        This meas ccw when y points down, and cw otherwise.
        In either case, rotating p[i+1]-p[i] by:
             90 degs * ori -> points outside
            -90 degs * ori -> points insdie
        """
        ## update orientation
        area = 0
        for ithis in range(0, self.points.shape[0]):
            inext = self.flatten_index( ithis + 1)
            dx = self.points[inext,0] - self.points[ithis,0]
            twoy = self.points[inext,1] + self.points[ithis,1]
            area = area + twoy * dx
        if(area >= 0): self.orientation = 1
        else         : self.orientation = -1

        ## Also update turn policy vector
        self.turn_policies = np.zeros(self.n_points(),dtype=int)
        for ii in range(0,self.n_points()):
            self.turn_policies[ii]  = self.turn_policy(ii)

    def update_corner2sample_map(self):
       """
       Inverse map for corners -> samples
       Works like the metis xadj or Compressed Row matrix formats.
       """
       assert( self.feature_points[0] == 0 )
        
       self.xcorner2sample = [0]
       icorner = 0
       for ifeature in range(0, self.n_feature_points() ):
           new_icorner = self.feature_points[ifeature]
           if( new_icorner != icorner ):
               icorner = icorner + 1
               self.xcorner2sample.append( ifeature )
       self.xcorner2sample.append( self.n_feature_points() )
       self.xcorner2sample = np.array( self.xcorner2sample, dtype=int)
       
    ## -------------------- Plotabble corner values
    def featurevalue2cornervalue(self, values, gather_operation):
        """
        The plot functions expect cornerscores to be have dimension of
        points. But corners and cscores have the dimensions of the
        feature points. This function takes care of this case:
        """
        assert( values.shape[0] == self.n_feature_points() )

        ## In this case values must be identical
        if(gather_operation is None):
            for i in range(0, self.n_points()):
                point_values = (
                    values[ self.xcorner2sample[i]:self.xcorner2sample[i+1] ])
                ## check values to be equal
                assert( np.all( np.isclose( point_values[0], point_values) ) )
                
            ## return value of the first sample for each point
            return values[ self.xcorner2sample[:-1] ]
        
        ## In this case values can be different
        else:
            answer = np.zeros( (self.n_points()) )
            for i in range(0, self.n_points()):
                point_values = (
                    values[ self.xcorner2sample[i]:self.xcorner2sample[i+1] ])
                answer[i] = gather_operation( point_values ) 
                
            ## gather all
            return answer
        
    ## -------------------- Normal at a point
    def normal(self, idx, ntype=0):
        """
        Calculate the normal at a point.
        ntype 0:
            p[idx+1] - p[idx], rotated such that it points outwards.
        """
        if(ntype==0):
            idx = self.flatten_index( idx )
            inext = self.flatten_index( idx + 1 )
            tang = self.points[inext] - self.points[idx]
            normal = np.array([-tang[1], tang[0]])
            normal = normal / np.linalg.norm(normal)
            return self.orientation * normal
        elif(ntype==1):
            iprev = self.flatten_index( idx - 1 )
            inext = self.flatten_index( idx + 1 )
            tang = self.points[inext] - self.points[iprev]
            normal = np.array([-tang[1], tang[0]])
            normal = normal / np.linalg.norm(normal)
            return self.orientation * normal
        else:
            assert(False)

    ## -------------------- Turn policy at a point
    def turn_policy(self, idx):
        """
        Return the turn policy at a point.
        1:          0:           -1:
            ---o       ---o---         |xxx
            xxx|       xxx xxx         |xxx
            xxx|       xxx xxx      ---oxxx
                                    xxx xxx
                                    xxx xxx
        """
        idx = self.flatten_index( idx )
        inext = self.flatten_index( idx + 1 )
        normal  = self.normal(idx-1,ntype=0)
        tang    = self.points[inext] - self.points[idx]

        ### round to make sure rounding of errors do not hurt
        ### dot as it is required
        ### sign to take care of non-unit length vectors
        ### int as the turn policy is an integer
        return -int( np.sign( np.round( np.dot( tang, normal ) ) ) )
    
    ## -------------------- Clone
    def clone(self):
        cl = Polygon()
        cl.name = copy.deepcopy(self.name)
        cl.resolution = copy.deepcopy(self.resolution)
        cl.points = copy.deepcopy(self.points)
        cl.closure_data = copy.deepcopy(self.closure_data)
        cl.feature_points = copy.deepcopy(self.feature_points)
        cl.corners = copy.deepcopy(self.corners)
        cl.point_data = copy.deepcopy(self.point_data)
        cl.orientation = copy.deepcopy(self.orientation)
        cl.turn_policies = copy.deepcopy(self.turn_policies)
        cl.xcorner2sample = copy.deepcopy(self.xcorner2sample)
        return cl
        
    ## -------------------- Mirror the polygon
    def mirror_points(self, isx,isy,isswap,isflip):

        if(isswap):
            self.points[:,[0,1]] =  self.points[:,[1,0]]
        if(isx):
            self.points[:,0] = -self.points[:,0]
        if(isy):
            self.points[:,1] = -self.points[:,1]
        if(isflip):
            self.points =  np.flip( self.points, 0)
            for data_name in self.point_data.keys():
                self.point_data[data_name] =  np.flip( self.point_data[data_name], 0)                
            self.corners =  np.flip( self.corners, 0)
            
            ## Update references to point indices ( sample points )
            self.feature_points = (self.n_points()-1) - self.feature_points
            self.feature_points = np.flip( self.feature_points , 0)
            ## Update references to point indices ( closure pairs )
            if(self.n_closure_candidates() > 0 ):
                mask = self.closure_data[:,:,0] >= 0
                self.closure_data[mask,0] = (
                    (self.n_points()-1) - self.closure_data[mask,0] )
                self.closure_data = np.flip( self.closure_data , 0)
                self.closure_data[:,:, [3,4,5,6]] = (
                    self.closure_data[:,:, [4,3,6,5]] )
            
        self.update_orientation()
        self.update_corner2sample_map()
        
    ## -------------------- Read the polygon from .txt file
    def read_polygon_txt(self, filename):
        ## reset the data
        self.init()
        self.name = filename.rsplit('/',2)[-2]
        
        reader = LineByLineReader(filename)
        xy = []

        ## start reading the coordinates
        while( reader.read_line() ):
            xy.append( reader.read_token( float ) )
            xy.append( reader.read_token( float ) )
        self.points = np.reshape(xy, (-1,2) )
        self.closure_data = np.zeros( (0,0,0) )
        self.corners = np.zeros(self.n_points(),dtype=int)-1
        self.feature_points = np.arange(0, self.n_points() )

        self.point_data = {}
        self.point_data[CORNER_SCORES] = np.zeros(self.n_points())-1
        
        self.update_orientation()
        self.update_corner2sample_map()
                
    ## -------------------- Read the corners from .txt file
    def read_corners_txt(self, filename):
        ## Make sure polygon is present
        assert( self.n_points() > 0 )
        assert( self.n_points() == self.n_feature_points() )

        ## reset the data
        self.corners = np.zeros( (self.n_points()), dtype=int ) - 1
        
        reader = LineByLineReader(filename)

        ## start reading the coordinates
        while( reader.read_line() ):
            cc = reader.read_token( int )
            if( cc < self.n_points() ):
                self.corners[ cc ] = 1

    ## -------------------- Read corner scores and the polygon
    ##                      alltogether
    def read_bundle(self, filename):
        
        data = np.loadtxt(filename,dtype=float)
        ## Read the file, first line must be polygon
        ## Address.
        self.init()
        self.name = filename
        self.points = data[:,[0,1]]
        self.corners = data[:,2]
        self.point_data = {}
        self.point_data[CORNER_SCORES] = data[:,3]
        self.update_orientation()
        self.feature_points = np.arange(0, self.n_points)
        self.update_corner2sample_map()
        self.closure_data = np.zeros( (0,0,0) )


    def write_bundle(self, filename):
        data = np.array([self.points[:,0],
                         self.points[:,1],
                         self.corners,
                         self.point_data[CORNER_SCORES]],
                        dtype=float)
        data = np.transpose(data)
        np.savetxt(filename, data)

    ## -------------------- Init from points
    def init_from_verts(self, verts):
        self.points = np.array(verts)
        self.feature_points = np.arange(0, self.n_points() )
        self.corners = np.zeros(self.n_feature_points(),dtype=int)-1
        self.update_orientation()
        self.update_corner2sample_map()
        
    ## -------------------- Read the new format
    def read_dot_poly(self, filename):
        "Read the new format. Everythin is included."
        reader = LineByLineReader(filename)

        ## Read the header
        reader.read_line()
        headerdict = reader.read_sandwich('{','}')
        self.name = headerdict['name']
        self.resolution = headerdict['resolution']
        n_points = headerdict['n_points']
        n_feature_points = headerdict['n_feature_points']
        n_closure_features = headerdict['n_closure_features']
        n_closure_candidates = headerdict['n_closure_candidates']
        has_corners = headerdict['has_corners']

        ##
        ## Read point data
        ## Backward compatibility, also support my older code
        ## which only supported corners scores as point data.
        ##
        n_point_data = None
        point_data_names = None
        if('has_corner_scores' in headerdict): ## Old format
            if(headerdict['has_corner_scores'] == True):
                n_point_data = 1
                point_data_names = [CORNER_SCORES]
            else:
               n_point_data = 0
               point_data_names = []
        else:                                  ## New format
            n_point_data = headerdict["n_point_data"]
            point_data_names = headerdict["point_data_names"]
            assert len(point_data_names) == n_point_data


        ##
        ## Start reading each case
        ##

        ## Points
        self.points = np.zeros( (n_points, 2) )
        for i in range(0, n_points):
            reader.read_line()
            self.points[i,0] = reader.read_token(float)
            self.points[i,1] = reader.read_token(float)

        ## feature points
        if(n_feature_points == -1):
            self.feature_points = np.arange(0, self.n_points() )
        else:
            self.feature_points = np.zeros(n_feature_points, dtype=int)
            for i in range(0, n_feature_points):
                reader.read_line()
                self.feature_points[i] = reader.read_token(int)

        ## corners
        self.corners = np.zeros(self.n_feature_points(),dtype=int)-1
        if(has_corners):
            assert( n_feature_points > 0 )
            for i in range(0, n_feature_points):
                reader.read_line()
                self.corners[i] = reader.read_token(float)

        ##
        ## Read Point data arrays
        ##
        point_data = []

        for data_name in point_data_names:
            data = np.zeros(self.n_points())
            point_data.append(data)
            self.point_data[data_name] = data
            
        if(n_point_data >= 0):
            for ipoint in range(0, self.n_points()):
                reader.read_line()
                for jdata in range(0, n_point_data):
                    point_data[jdata][ipoint]= reader.read_token(float)

        ## closure_data -- probably remove
        if(n_closure_features <= 0):
            self.closure_data = np.zeros( (0,0,0) )
        else:
            assert( n_closure_candidates > 0 )
            self.closure_data = np.zeros(
                (self.n_points(),
                 n_closure_candidates,
                 n_closure_features), dtype=float )
            for i in range(0, n_points):
                for j in range(0, n_closure_candidates):
                    reader.read_line()
                    i_sanity = reader.read_token(int)
                    assert( i == i_sanity )
                    for k in range(0, n_closure_features):
                        self.closure_data[i,j,k] = reader.read_token(float)
            
        ## Update orientation
        self.update_orientation()
        self.update_corner2sample_map()
        
    def write_dot_poly(self, filename):
        "Write the new format."
        ofile = open(filename, 'w')
        
        reader = LineByLineReader(filename)

        ## Read the header 
        headerdict = {}
        headerdict['name'] = self.name
        headerdict['resolution'] = self.resolution 
        headerdict['n_points'] = self.n_points() 
        headerdict['n_feature_points'] = self.n_feature_points() 
        headerdict['n_closure_features'] = self.n_closure_features() 
        headerdict['n_closure_candidates'] = self.n_closure_candidates() 
        headerdict['n_point_data'] = len(self.point_data)
        headerdict['point_data_names'] = list(self.point_data.keys())
        headerdict['has_corners'] = True
        ofile.write('%s \n\n' % (headerdict) )

        ##
        ## Start writing each case
        ##

        ## Points
        ofile.write('## Point coordiates\n')
        for i in range(0, self.n_points()):
            ofile.write('%f %f\n' % (self.points[i,0], self.points[i,1]) )
        ofile.write('\n')
            
        ## Feature points
        ofile.write('## Feature points (for handling label disputes'
                     'between artists) \n')
        for i in range(0, self.n_feature_points()):
            ofile.write('%d\n' % self.feature_points[i])
        ofile.write('\n')

        ## corners
        ofile.write('## Artist corner annotations'
                     '(All -1 if unavailable) \n')
        for i in range(0, self.n_feature_points()):
            ofile.write('%f\n'% self.corners[i] )
        ofile.write('\n')        
            
        ## point data
        ## First put into numpy array and then dump array into file
        ofile.write('## Point data \n')
        point_data = []
        for point_data_name in self.point_data.keys():
            point_data.append(self.point_data[point_data_name])
        point_data = np.transpose(np.array(point_data))
        np.savetxt(ofile, point_data)
        ofile.write('\n')        

        ## closure_data
        ofile.write('## Closure data. The format is: \n'
                     '# point idx, <candidate1 idx> <metric1> <metric2> ...\n'
                     '# point idx, <candidate2 idx> <metric1> <metric2> ...\n')
        if(self.n_closure_features() > 0):
            for i in range(0, self.n_points()):
                for j in range(0, self.n_closure_candidates()):
                    ofile.write('%d '% i)
                    for k in range(0, self.n_closure_features()):
                        ofile.write('%f '% self.closure_data[i,j,k] )
                    ofile.write('\n')

        ofile.close()

    ## -------------------- Number of points
    def n_points(self):
        assert (self.points.shape[1] == 2) 
        return self.points.shape[0]
    
    def n_feature_points(self):
        # assert (self.feature_points.shape[1] == 1) 
        return self.feature_points.shape[0]

    def is_closure_data_present(self):
        return ( self.closure_data.shape[0] == self.n_points() )

    def n_closure_features(self):
        return ( self.closure_data.shape[2] )

    def n_closure_candidates(self):
        return ( self.closure_data.shape[1] )

    ## -------------------- Convert circular index to flat
    def flatten_index(self, index):
        index = index + 3 * self.n_points()
        return index % self.n_points()

    ## -------------------- Convert circular index to flat
    def is_feature_point_corner(self, index):
        assert ( (index >= 0) and (index < self.n_feature_points()) )
        return ( np.abs(self.corners[index] - 1) < 1e-10 )
        
    ## -------------------- FInd the predicted corners
    ## Depracated
    def get_corner_scores(self, regressor, cdi_options):
        assert(0)
        """
        TODO: BATCH THE EVALUTAITON STEP.
        """
        self.corner_scores = []
        
        for ii in range(0, self.n_points() ):
            cdi = CornerDetectionInput(self, ii, cdi_options)
            self.corner_scores.append(
                cdi.get_regress_score( regressor ) )

        self.corner_scores = np.array(self.corner_scores)
        self.corner_scores.reshape(-1,1)
        return self.corner_scores

    ##---------------------- A very nice plot
    def plot_nice(self,ax,corner_scores=None,**kwargs):
        reverse_y_axis = kwargs.pop('reverse_y',True)
        plot_cbar = kwargs.pop('plot_cbar',False)
        plot_ori = kwargs.pop('plot_ori',False)
        s_corners = kwargs.pop('s_corners',0.5)
        lw = kwargs.pop('lw',2)
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))

        ## Plot the polygon and the corners
        if( corner_scores is not None):
            self.plot_polygon( ax,
                               corner_scores=corner_scores,
                               options_plt=dict(lw=lw,zorder=1))
            self.plot_corners( ax,
                               corner_scores=corner_scores,
                               options_plt=dict(s=s_corners,zorder=0.5))
        else:
            self.plot_polygon( ax,
                               options_plt=dict(lw=lw,zorder=1))
            
        ## Other stuff
        self._plot_config_ax(ax, reverse_y_axis)
            
        ## Create the color map
        
        if(plot_cbar):
            norm = matplotlib.colors.Normalize(vmin=-1,vmax=1)
            s_m = matplotlib.cm.ScalarMappable(cmap=_cmap_of_choice, norm=norm)
            s_m.set_array([])
            fig = ax.get_figure()
            fig.colorbar(s_m,orientation='vertical',ticks=[-1,0,1])
            
 
        ## Remove the tikz
        # ax.get_xaxis().set_visible(False)
        # ax.get_yaxis().set_visible(False)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        
        
        ## Write the orientation (ccw, or cw)
        if(plot_ori):
            ax.text(0.5,0.5,'{0}'.format(self.orientation),
                    transform=ax.transAxes, fontsize=10)

        
    ## -------------------- Setup the axis
    def _plot_config_ax(self, ax, reverse=True):
        ## Set the limits
        limits = 1 
        ax.set_xlim(0, limits)
        if(reverse): ax.set_ylim(limits, 0)
        else:        ax.set_ylim(0, limits)
        ax.set_aspect('equal')
        ax.axis('equal')

    ## -------------------- Plot the polygon
    def plot_polygon(self,ax,**kwargs):
        """
        Plot the polygonal.
        If supplied, plot the corner probabilities as well.
        """

        ##
        ## Read the options
        ##
        is_closed = kwargs.pop('is_closed', True)
        corner_scores = kwargs.pop('corner_scores', None)
        cmapname = kwargs.pop('cmap','cool')
        options_plt = kwargs.pop('options_plt',dict())
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))
        
        ##
        ## If corner values are not given, simply draw the
        ## polygon.
        ##
        if( type(corner_scores) == type(None) ):
            if(is_closed):
                xx = np.append(self.points[:,0], self.points[0,0])
                yy = np.append(self.points[:,1], self.points[0,1])
            else:
                xx = self.points[:,0]
                yy = self.points[:,1]
            ax.plot(xx,yy,**options_plt)
        ##
        ## Otherwise, assign colors to the half segments after, and
        ## before each line segment.
        ##
        else:
            ##
            ## Create the segments
            ##
            segments = []
            for ip in range(0,self.n_points()):
                ## Special case 1:
                if( (not is_closed) and (ip == 0) ):
                    assert(0 and "Not supported")
                    pass
                ## Special case 1:
                elif( (not is_closed) and (ip == n_points()-1) ):
                    assert(0 and "Not supported")
                    pass
                ## Usual case
                else:
                    iprev = ((ip-1)+self.n_points()) % self.n_points()
                    inext = ((ip+1)+self.n_points()) % self.n_points()
                    p1 = 0.5 * (self.points[iprev] + self.points[ip])
                    p2 = self.points[ip]
                    p3 = 0.5 * (self.points[ip] + self.points[inext])
                    seg  = []
                    seg.append(p1)
                    seg.append(p2)
                    seg.append(p3)
                    segments.append(seg)
            ##
            ## Create line collection, and  Assign colors
            ## to them
            ##
            ## find the min and max value
            normalscores = corner_scores.copy()

            # eps = 1e-7
            # normalscores[ normalscores > 1] = 1
            # normalscores[ normalscores < -1] = -1
            # normalscores[np.logical_and(
            #     (normalscores < eps),  (normalscores >= 0)) ] = eps
            # normalscores[np.logical_and(
            #     (normalscores > -eps), (normalscores <= 0)) ] = -eps

            #  maskp = (normalscores > 0)
            # maskm = (normalscores < 0)

            # normalscores[ maskp ] = normalscores[ maskp ]
            # normalscores[ maskm ] = -normalscores[ maskm ]**2
                        
            vmin=-1
            vmax=+1
            # signs = np.sign(corner_scores)
            # mappedscores=np.multiply(
            #     np.cbrt(normalscores-0.5), signs) + 0.5
            segments = np.array(segments)
            #_cmap = plt.get_cmap(cmapname)
            lc = LineCollection(segments, cmap=_cmap_of_choice,
                                 norm=matplotlib.colors.Normalize(vmin,vmax))
            lc.set_array(normalscores) 
            
            ##
            ## plot the collection
            ##
            ax.add_collection(lc)

        ##
        ## Write the orientation (ccw, or cw)
        ## And configure axis
        ##
        #ax.text(0.5,0.5,'{0}'.format(self.orientation)
        #        ,transform=ax.transAxes, fontsize=10)

    ## -------------------- Annotate the corners
    def annotate_corners(self, ax, **kwargs):
        """
        Annotate the corners
        """
        options_plt = kwargs.pop('options_plt',dict())
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))
        
        ## Default value for font size
        psize = options_plt.pop('size',12)
        options_plt['size'] = 12

        ## draw
        pts = self.points
        for i in range(0, self.points.shape[0]):
            ax.annotate(str(i), pts[i]+[0.1,0.1], **options_plt )

    ## -------------------- Plot dots on corners
    def plot_corners(self,ax,**kwargs):
        """
        Plot dots on the polygon corners
        """
        corner_scores = kwargs.pop('corner_scores', None)
        cmapname = kwargs.pop('cmap','bwr')
        options_plt = kwargs.pop('options_plt',dict())
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))

        xx = self.points[:,0]
        yy = self.points[:,1]

        ##
        ## Just draw, when there are no corner scores
        ##
        if( corner_scores is None ):
            ax.scatter(xx, yy,**options_plt)
        ##
        ## Otherwise, find the colors:
        ##
        else:
            normalscores = corner_scores

            eps = 1e-7
            normalscores[ normalscores > 1] = 1
            normalscores[ normalscores < -1] = -1
            # normalscores[np.logical_and(
            #     (normalscores < eps),  (normalscores >= 0)) ] = eps
            # normalscores[np.logical_and(
            #     (normalscores > -eps), (normalscores <= 0)) ] = -eps

            # maskp = (normalscores > 0)
            # maskm = (normalscores < 0)

            # normalscores[ maskp ] = normalscores[ maskp ]**2
            # normalscores[ maskm ] = -normalscores[ maskm ]**2

            vmin = -1
            vmax = +1
            smap = matplotlib.cm.ScalarMappable(
                norm=matplotlib.colors.Normalize(vmin,vmax),
                cmap=_cmap_of_choice)
            colors = smap.to_rgba(normalscores)
            mask = corner_scores < -.9
            colors[mask,3]  = 0.
            ## Set the colors and draw
            options_plt['color'] = colors
            ax.scatter(xx, yy,**options_plt)

    ## -------------------- Annotate turn policy
    def annotate_turn_policy(self,ax,**kwargs):
        options_plt = kwargs.pop('options_plt',dict())
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))

        ##
        ## Loop over vertices and annotate
        ##
        colors = ['r','k','b']
        for ip in range(0,self.n_points()):
            point = self.points[ip]
            # normal  = self.normal(ip,ntype=1)
            tp = self.turn_policies[ip]
            # ax.annotate(str(tp), point+0.1*normal, **options_plt )
            ax.scatter(point[0],point[1],c=colors[tp+1],**options_plt)

    ## -------------------- Plot Normals
    def plot_normals(self,ax,**kwargs):    
        options_plt = kwargs.pop('options_plt',dict())
        ntype = kwargs.pop('ntype',0)
        if kwargs:
            raise ValueError("Extra values given {0}".format( kwargs ))

        for ip in range(0,self.n_points()):
            p0 = self.points[ip]
            p1  = 0.5*self.normal(ip,ntype=ntype) + p0
            ax.plot([p0[0], p1[0]], [p0[1], p1[1]], **options_plt)


    def get_reso(self):
        pmax = np.max( self.points, 0)
        pmin = np.min( self.points, 0)
        diff = pmax - pmin
        dim = np.max( diff )
        l2 = np.log2(dim)
        l2 = np.round(l2)
        reso = 2**l2
        reso =  int(reso)
        reso = np.max( [reso, 32] )
        reso = np.min( [reso, 256] )
        return reso
