#!/usr/bin/env python3

# https://stackoverflow.com/questions/19966056/pyqt5-how-can-i-connect-a-qpushbutton-to-a-slot

import sys

from core.manager import Manager
from core.defs import *

from PyQt4 import QtGui, QtCore
import matplotlib
matplotlib.use(MPL_BACKEND)

from qtgui.canvas_widget import CanvasWidget
from qtgui.button_list_widget import ButtonListWidget


BUTTON_READ_IMAGE_COLOR      = "BUTTON_READ_IMAGE_COLOR"
BUTTON_READ_IMAGE_BW         = "BUTTON_READ_IMAGE_BW"
BUTTON_INVERT_IMAGE    = "BUTTON_INVERT_IMAGE"
BUTTON_CROP_IMAGE      = "BUTTON_CROP_IMAGE"
BUTTON_RESET_IMAGE     = "BUTTON_RESET_IMAGE"
BUTTON_EXTRACT_POLY    = "BUTTON_EXTRACT_POLY"
BUTTON_PREDICT         = "BUTTON_PREDICT"
BUTTON_FIT             = "BUTTON_FIT"
BUTTON_SAVE_IMAGE      = "BUTTON_SAVE_IMAGE"
BUTTON_SHOW_COLOR      = "BUTTON_SHOW_COLOR"

BUTTON_ACCEPT          = "BUTTON_ACCEPT"
BUTTON_REFUSE          = "BUTTON_REFUSE"

class MainWindow(QtGui.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle('Perception-Driven Vectorization')
        self._is_gui_setup = False
        ##
        self.manager = Manager()
        self.manager.add_observer(self)
        self.setup_gui()
        ##
        self.setMinimumSize(1200, 900)
        self.setMinimumSize(600, 450)
        # self.canvas.resize(1100, 900);
        
    def setup_gui(self):
        if(self._is_gui_setup): return
        ## Layout
        self.setLayout( QtGui.QHBoxLayout(self) )
        ## Control panel and canvas
        self.control_panel = ButtonListWidget(self, is_vertical=True)
        self.canvas        = CanvasWidget(self)
        self.layout().addWidget( self.canvas, 5 )
        self.layout().addWidget( self.control_panel, 1 )

        ## Add buttons
        self.control_panel.add_button(BUTTON_READ_IMAGE_BW,  "Read Single Region")
        self.control_panel.add_button(BUTTON_READ_IMAGE_COLOR,  "Read Multi Color")
        self.control_panel.add_button(BUTTON_FIT,         "Vectorize")
        self.control_panel.add_button(BUTTON_SHOW_COLOR,  "(Experimental) Show Colors")
        self.control_panel.add_button(BUTTON_SAVE_IMAGE,  "Save")

        b = self.control_panel.get_button(BUTTON_READ_IMAGE_COLOR)
        b.clicked.connect(self.read_image_clicked_color)
        #
        b = self.control_panel.get_button(BUTTON_READ_IMAGE_BW)
        b.clicked.connect(self.read_image_clicked_bw)
        # ##
        b = self.control_panel.get_button(BUTTON_FIT)
        b.clicked.connect(self.fit_clicked)

        b = self.control_panel.get_button(BUTTON_SAVE_IMAGE)
        b.clicked.connect(self.save_image_clicked)

        b = self.control_panel.get_button(BUTTON_SHOW_COLOR)
        b.clicked.connect(self.show_color_clicked)
        
    ### =================================================
    ###  CLICKS FROM THE BUTTONS
    ### =================================================
    def read_image_clicked_color(self):
        filename = QtGui.QFileDialog.getOpenFileName(self, 'OpenFile')
        if len(filename) > 0:
            self.manager.read_image(filename, 1)

    def read_image_clicked_bw(self):
        filename = QtGui.QFileDialog.getOpenFileName(self, 'OpenFile')
        if len(filename) > 0:
            self.manager.read_image(filename, 0)

    def fit_clicked(self):
        self.manager.run_fitter()

    def save_image_clicked(self):
        filename = QtGui.QFileDialog.getSaveFileName(self, 'SaveFile')
        self.manager.save_image(filename)

    def show_color_clicked(self):
        if self.manager.svg_address is None:
            print('You need to vectorize something first!')
        self.manager.toggle_view_colors()
        if self.manager.view_colors == 0:
            self.control_panel.get_button(BUTTON_SHOW_COLOR).setText('(Experimental) Show Colors')
        else:
            self.control_panel.get_button(BUTTON_SHOW_COLOR).setText('Show Black & White')

    ### =================================================
    ###  THINGS TO OBSERVE FROM MANAGER
    ### =================================================
    def observe_image_changed(self, manager):
        if( manager.to_view == VIEW_NONE ):
            self.canvas.clear_figure()
        elif( manager.to_view == VIEW_PNG  ):
            matplotlib.pyplot.figure(0)
            matplotlib.pyplot.clf()
            if(manager.is_multicolor):
                matplotlib.pyplot.imshow(manager.image)
            else:
                cmap = matplotlib.pyplot.get_cmap('gray')
                matplotlib.pyplot.imshow(manager.image, cmap=cmap)
            matplotlib.pyplot.gca().get_xaxis().set_visible(False)
            matplotlib.pyplot.gca().get_yaxis().set_visible(False)
            self.canvas.show_mpl( matplotlib.pyplot.gcf() )
        elif( manager.to_view == VIEW_POLY  ):
            matplotlib.pyplot.figure(0)
            matplotlib.pyplot.clf()
            matplotlib.pyplot.gca().get_xaxis().set_visible(False)
            matplotlib.pyplot.gca().get_yaxis().set_visible(False)
            self.manager.polygon.plot_nice( matplotlib.pyplot.gca() )
            self.canvas.show_mpl( matplotlib.pyplot.gcf() )
        elif( manager.to_view == VIEW_SCORES  ):
            scores =self.manager.polygon.point_data['y_proba']
            # print(scores)
            matplotlib.pyplot.figure(0)
            matplotlib.pyplot.clf()
            matplotlib.pyplot.gca().get_xaxis().set_visible(False)
            matplotlib.pyplot.gca().get_yaxis().set_visible(False)
            self.manager.polygon.plot_nice( matplotlib.pyplot.gca(), corner_scores=scores )
            self.canvas.show_mpl( matplotlib.pyplot.gcf() )            
        elif( manager.to_view == VIEW_SVG  ):
            self.canvas.show_svg( manager.svg_address )            
        else:
            self.canvas.clear_figure()    

        
class DisclaimerWindow(QtGui.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle('Disclaimer')
        self.accepted = False
        self.setMinimumSize(600, 450)

        self.setLayout( QtGui.QVBoxLayout(self) )

        self.disclaimer = QtGui.QTextEdit()
        self.control_panel = ButtonListWidget(self, is_vertical=False)

        self.layout().addWidget(self.disclaimer, 6)
        self.layout().addWidget( self.control_panel, 1 )

        self.control_panel.add_button(BUTTON_ACCEPT,  "Accept")
        self.control_panel.add_button(BUTTON_REFUSE,  "Decline")

        b = self.control_panel.get_button(BUTTON_ACCEPT)
        b.clicked.connect(self.accept_clicked)
        
        b = self.control_panel.get_button(BUTTON_REFUSE)
        b.clicked.connect(self.refuse_clicked)

        with open(EULA_DIR, 'rb') as ff:
            eula = ff.read()
            eula = eula.decode('windows-1252')

        self.disclaimer.insertHtml(eula)
        self.disclaimer.moveCursor(QtGui.QTextCursor.Start)
        self.disclaimer.setReadOnly(True)

    def accept_clicked(self):
        self.accepted = True
        self.close()

    def refuse_clicked(self):
        self.close()


## ====================================================
##                     Test code
## ====================================================
# Create an instance of the application window and run it
def main():
    qt_app = QtGui.QApplication(sys.argv)
    qt_app.setApplicationName('Raster to Vector')

    dw = DisclaimerWindow()
    dw.exec_()

    if not dw.accepted:
        print('Disclaimer not accepted')
        sys.exit(1)

    m = MainWindow()
    m.show()

    print("Starting the GUI")
    sys.exit(qt_app.exec_())

if __name__ == '__main__':
    main()
